import sys, json, os
import sayuDB.processor as sayuDB

with open(f'{os.path.dirname(__file__)}/users.json', 'r') as user:
    user = json.load(user)
with open(f'{os.path.dirname(__file__)}/config.json', 'r') as config:
    config = json.load(config)

def save_conf():
    with open(f'{os.path.dirname(__file__)}/users.json', 'w')as wconf:
        json.dump(user, wconf, indent=4)
    with open(f'{os.path.dirname(__file__)}/config.json', 'w')as wconf:
        json.dump(config, wconf, indent=4)
    return

try:
    sys.argv[1]
except:
    print("Read the doc here : https://github.com/Arsybai/blob/main/README.md")
    print("""Invalid commad. get help by
--h""")
    exit()

if sys.argv[1] == "help" or sys.argv[1] == '--h':
    print("sayuDB v0.0.6\nRead the doc here : https://github.com/Arsybai/blob/main/README.md")
    print("""
[ USERS ]
Show Users  \t\t: show users
Create user \t\t: create user <username> <password>
Remove user \t\t: remove user <username>

[ DATABASE ]
Show Database   \t: show databases
Create Database \t: create database <database_name>
Drop Database \t\t: drop database <database_name>
Grant user \t\t: grant user <username> <database_name>
Import Database \t: import database <path_to_ezdb>
Export Database \t: export database <database_name> <output_path>

[ SERVER ]
Activate Server \t: activate server <port>
Deactivate Server \t: deactivate server
Block IP  \t\t: block ip <ip> (coming soon)
Unblock IP  \t\t: unblock ip <ip> (coming soon)

if you can't access the server after activate. setup the public access by
setup public server <server_name> <port>
leave it with your server public IP in server_name if you don't have domain and the port must be same when you activate
after that you can acces it within your IP server.
    """)

elif sys.argv[1] == 'create':
    if sys.argv[2] == 'user':
        username_ = sys.argv[3]
        if username_ in user:
            sys.exit("Username already exist")
        password_ = sys.argv[4]
        user[username_] = {
            "username": username_,
            "password": password_,
            "access": []
        }
        save_conf()
        print("User created")
    elif sys.argv[2] == 'database':
        sayuDB.create_database(sys.argv[3])

elif sys.argv[1] == 'show':
    if sys.argv[2] == 'users':
        for i in user:
            print(f"> {i}")
    elif sys.argv[2] == 'databases':
        sayuDB.show_databases()

elif sys.argv[1] == 'remove':
    if sys.argv[2] == 'user':
        if sys.argv[3] in user:
            del user[sys.argv[3]]
            print("User deleted")
            save_conf()
        else:
            print("User not found")

elif sys.argv[1] == 'drop' and sys.argv[2] == 'database':
    sayuDB.drop_database(sys.argv[3])

elif sys.argv[1] == 'grant' and sys.argv[2] == 'user':
    if sys.argv[3] not in user:
        exit("User not found")
    if sys.argv[4] not in sayuDB.show_databases():
        exit("Database not found")
    user[sys.argv[3]]['access'].append(sys.argv[4])
    save_conf()
    print(f"Granted all privileges of {sys.argv[4]} to {sys.argv[3]}")

elif sys.argv[1] == 'activate' and sys.argv[2] == 'server':
    import os, platform
    if platform.system() == 'Windows':
        os.system(f'py {os.path.dirname(__file__)}/server.py {sys.argv[3]}')
    else:
        os.system(f"screen -dmS sayuDB")
        os.system(f"screen -S sayuDB -X stuff 'python3 {os.path.dirname(__file__)}/server.py {sys.argv[3]}\n'")
    print("Server activated. access it within your IP")

elif sys.argv[1] == 'deactivate' and sys.argv[2] == 'server':
    import os, platform
    if platform.system() == 'Windows':
        os.system(f'Just press ctrl+c')
    else:
        os.system(f"screen -X -S sayuDB kill")
    print("Server Deactivated.")

elif sys.argv[1] == 'import' and sys.argv[2] == 'database':
    sayuDB.import_database(sys.argv[3])
elif sys.argv[1] == 'export' and sys.argv[2] == 'database':
    sayuDB.export_database(sys.argv[3], sys.argv[4])

elif sys.argv[1] == 'setup' and sys.argv[2] == 'public':
    if sys.argv[3] == 'server':
        nginx_conf = """server {
    listen 80;
    server_name """+sys.argv[4]+""";
    location / {
        proxy_pass http://127.0.0.1:"""+sys.argv[5]+""";
        proxy_set_header Host $host;
        proxy_set_header X-Forwarded-For $proxy_add_x_forwarded_for;
    }
}"""
        os.system("sudo apt install nginx -y")
        os.system("sudo ufw allow 'Nginx Full'")
        with open("/etc/nginx/sites-enabled/sayuDB", "w") as wcc:
            wcc.write(nginx_conf)
        os.system("sudo systemctl restart nginx")
        os.system("clear")
        print(f"Successfully setup public server.\nhttp://{sys.argv[4]}")
    else:
        print("Public what? what must I setup?")


import json
import requests
import os
import base64

def encode_base64(input_string):
    encoded_bytes = base64.b64encode(input_string.encode('utf-8'))
    encoded_string = encoded_bytes.decode('utf-8')
    return encoded_string

def decode_base64(encoded_string):
    decoded_bytes = base64.b64decode(encoded_string.encode('utf-8'))
    decoded_string = decoded_bytes.decode('utf-8')
    return decoded_string

def check_network(host):
    response = requests.get(host)
    return response

def sendRequest(host, data, method='GET'):
    if method == 'GET':
        response = requests.get(host, data=data).text
    elif method == 'POST':
        response = requests.post(host, data=data).text
    else:
        response = "Invalid method"
    return response

class remote:

    def __init__(self, username:str, password:str, host:str):
        self.username = username
        self.password = password
        self.host = host
        response = sendRequest(host+'/oauth', {'username':encode_base64(username), 'password':encode_base64(password)})
        if response != 'ok':
            exit(response)
        return
    
    def database_exsistence(self, database:str):
        response = sendRequest(self.host+'/existence', {'username':encode_base64(self.username), 'password':encode_base64(self.password), 'database':database})
        if response != 'ok':
            exit(response)
        return True
    
    def commit_database(self, database:str, content):
        response = sendRequest(self.host+'/commit', {'username':encode_base64(self.username), 'password':encode_base64(self.password), 'database':database, 'content':str(encode_base64(content))}, 'POST')
        if response != 'ok':
            exit(response)
        return True
    
    def pull_database(self, database:str):
        response = sendRequest(self.host+'/pull', {'username':encode_base64(self.username), 'password':encode_base64(self.password), 'database':database})
        return json.loads(response)


import json, os, shutil, sys
from tabulate import tabulate
from operator import itemgetter
import sayuDB.remote as remote

def create_database(name: str):
    config_ = {
        "name":name
    }
    if not os.path.isfile(f'{os.path.dirname(__file__)}/datas/{name}.ezdb'):
        with open(f'{os.path.dirname(__file__)}/datas/{name}.ezdb', 'w') as cdb:
            json.dump(config_, cdb, indent=4)
        print(f"Database created [ {name} ]")
    else:
        print("Database already exist")
        
def show_databases():
    databases = os.listdir(f'{os.path.dirname(__file__)}/datas/')
    lists = []
    for i in databases:
        if '.ezdb' in i:
            print(f"> {i.replace('.ezdb','')}")
            lists.append(i.replace('.ezdb',''))
    return lists
        
def drop_database(name: str):
    if os.path.isfile(f'{os.path.dirname(__file__)}/datas/{name}.ezdb'):
        os.remove(f'{os.path.dirname(__file__)}/datas/{name}.ezdb')
        print(f"Database dropped [ {name} ]")
    else:
        print("Database not found")
        
def export_database(name: str, path_: str):
    if os.path.isfile(f'{os.path.dirname(__file__)}/datas/{name}.ezdb'):
        shutil.copyfile(f'{os.path.dirname(__file__)}/datas/{name}.ezdb', f"{path_}/{name}.ezdb")
        print(f"Database exported [ {name}.ezdb ]")
    else:
        print("Database not found")
        
def import_database(path_: str):
    if os.path.isfile(path_):
        with open(path_, 'r') as rdb:
            try:
                db_ = eval(rdb.read())
            except:
                db_ = rdb.read()
        with open(f'{os.path.dirname(__file__)}/datas/{db_["name"]}.ezdb', 'w') as wdb:
            try:
                json.dump(db_, wdb, indent=4)
                print(f"Database imported [ {db_['name']} ]")
            except:
                print("Import error")
    else:
        print("File or directory not found")
        

class sayuDB:
    
    def __init__(self, database, username:str=None, password:str=None, host:str=None):
        self.database = database
        self.is_remote = False
        if username != None and host != None:
            self.is_remote = True
            self.remote = remote.remote(username, password, host)
            self.remote.database_exsistence(database)
        return None
    
    def save_table(self, name: str, content: object):
        if self.is_remote:
            self.remote.commit_database(self.database, content)
        else:
            with open(f'{os.path.dirname(__file__)}/datas/{self.database}.ezdb', 'w') as wdb:
                json.dump(content, wdb, indent=4)
        return

    def save_db(self, content: object):
        with open(f'{os.path.dirname(__file__)}/datas/{self.database}.ezdb', 'w') as wdb:
            json.dump(content, wdb, indent=4)
        return
    
    def openDB(self):
        if self.is_remote:
            db_ = self.remote.pull_database(self.database)
        else:
            with open(f'{os.path.dirname(__file__)}/datas/{self.database}.ezdb', 'r') as rdb:
                db_ = json.loads(rdb.read())
        return db_
    
    def eval_typedata(self, typedata, content):
        try:
            return eval(f'{typedata}({content})')
        except:
            return content
    
    def create_table(self, name: str, col: list):
        """_summary_
        
        Args:
            name (str): the table name
            col (list): [['column_name', 'datatype']]
        
        the datatypes:
        - str   
        - float
        - int   
        - dict
        """
        with open(f'{os.path.dirname(__file__)}/datas/{self.database}.ezdb', 'r') as rdb:
            rdb = eval(rdb.read())
        if name in rdb:
            print("Table already exist")
        else:
            tables_ = {
                "name": name,
                "column": {},
                "datas":[]
            }
            for i in col:
                tables_["column"][i[0]] = i[1]
            rdb[name] = tables_
            self.save_table(name, rdb)
            print("Table created")
            
    def drop_table(self, name: str):
        with open(f'{os.path.dirname(__file__)}/datas/{self.database}.ezdb', 'r') as rdb:
            rdb = eval(rdb.read())
        if name not in rdb:
            print("Table not found")
        else:
            del rdb[name]
            self.save_table(name, rdb)
            print("Table dropped")
            
    def show_table(self, name: str):
        db_ = self.openDB()
        if name not in db_:
            print("Table not found")
        else:
            headeer_ = []
            rows_ = []
            for i in db_[name]['column']:
                headeer_.append(i)
            for i in db_[name]['datas']:
                row_ = []
                for u in i:
                    row_.append(i[u])
                rows_.append(row_)
            print(tabulate(rows_, headers=headeer_))
            
    def insert_row(self, table: str, col, contents: list):
        """_summary_

        Args:
            table (str): the table name
            col (list): column name. EX ['col1', 'col2']
            contents (list): the content for each column EX ['content 1', 'content 2']
            
            _Note : You can also use str for col with comma for the separated EX 'col1,col2'_
        """
        db_ = self.openDB()
        if type(col) == str:
            col = col.split(',')
        idx = 0
        data_ = {}
        for i in col:
            if i not in db_[table]['column']:
                print(f"Column {i} not found")
                sys.exit()
            elif len(col) != len(contents):
                print("The length of column and contents doesn't match")
                sys.exit()
            elif db_[table]['column'][i] == 'str':
                try:
                    data_[i] = str(contents[idx])
                except:
                    print(f"The content typedata invalid at column {i} content {contents[idx]}")
                    sys.exit()
            elif db_[table]['column'][i] == 'int':
                try:
                    data_[i] = int(contents[idx])
                except:
                    print(f"The content typedata invalid at column {i} content {contents[idx]}")
                    sys.exit()
            elif db_[table]['column'][i] == 'float':
                try:
                    data_[i] = float(contents[idx])
                except:
                    print(f"The content typedata invalid at column {i} content {contents[idx]}")
                    sys.exit()
            elif db_[table]['column'][i] == 'dict':
                try:
                    try:
                        data_[i] = eval(contents[idx])
                    except:
                        data_[i] = eval(str(contents[idx]))
                except:
                    print(f"The content typedata invalid at column {i} content {contents[idx]}")
                    sys.exit()
            idx += 1
        db_[table]['datas'].append(data_)
        self.save_table(table, db_)
        
    def select_row(self, table: str, col, where=None, order_by=None, as_json=False, limit=None):
        """_summary_

        Args:
            table (str): The table name
            col (_type_): The column what you want to show. use * for all column
            where (_type_, optional): where the column contain some value. Defaults to None.
            order_by (_type_, optional): short data by column asc or desc. Defaults to None.
            as_json (bool, optional): return as json or table. Defaults to False.
        """
        db_ = self.openDB()
        if limit != None:
            datas_ = db_[table]['datas'][:limit]
        else:
            datas_ = db_[table]['datas'][:limit]
        if col == '*':
            datas_ = datas_
        else:
            col = col.split(',')
            idx = 0
            for i in datas_:
                for u in list(i):
                    if u not in col:
                        del i[u]
                idx += 1
        if where != None and ' contain ' in where:
            temp_ = []
            for i in datas_:
                if self.eval_typedata('str', where.split(" contain ")[1]) in str(i[where.split(" contain ")[0]]):
                    temp_.append(i)
            datas_ = temp_
        if where != None and ' contain ' not in where:
            temp_ = []
            for i in datas_:
                if i[where.split("=")[0]] == self.eval_typedata(db_[table]['column'][where.split("=")[0]], where.split("=")[1]):
                    temp_.append(i)
            datas_ = temp_
        
        if order_by != None:
            key_ = order_by.split("|")[0]
            sort_ = order_by.split("|")[1]
            if sort_ == 'asc':
                datas_ = sorted(datas_, key=itemgetter(key_))
            elif sort_ == 'desc':
                datas_ = sorted(datas_, key=itemgetter(key_), reverse=True)
                    
        #the anu anu
        if as_json:
            try:
                return json.loads(datas_)
            except:
                return datas_
        headeer_ = []
        rows_ = []
        try:
            if datas_ != []:
                for i in datas_[0]:
                    headeer_.append(i)
            elif col == '*':
                for i in db_[table]['column']:
                    headeer_.append(i)
            else:
                for i in col:
                    headeer_.append(i)
        except:
            for i in db_[table]['column']:
                headeer_.append(i)
        for i in datas_:
            row_ = []
            for u in i:
                row_.append(i[u])
            rows_.append(row_)
        return f"{tabulate(rows_, headers=headeer_)}\n\n{len(rows_)} Rows"
    
    def update_row(self, table: str, set_: str, where: str):
        db_ = self.openDB()
        if table not in db_:
            sys.exit("Table not found")
        if ' contain ' in where:
            col_ = where.split(" contain ")[0]
            val_ = where.split(" contain ")[1]
        else:
            col_ = where.split("=")[0]
            val_ = where.split("=")[1]
        set_col = set_.split("=")[0]
        set_val = set_.split("=")[1]
        if col_ not in db_[table]['column']:
            sys.exit("Column not found")        
        if set_col not in db_[table]['column']:
            sys.exit("Column not found")        
        for i in db_[table]['datas']:
            if 'contain' in where:
                if str(val_) in str(i[col_]):
                    i[set_col] = set_val
            else:
                if i[col_] == val_:
                    i[set_col] = set_val
        self.save_table(table, db_)
        return "Column updated"
    
    def delete_row(self, table: str, where: str):
        db_ = self.openDB()
        if table not in db_:
            sys.exit("Table not found")
        if '=' in where:
            col_ = where.split('=')[0]
            val_ = where.split('=')[1]
            idx = 0
            for i in db_[table]['datas']:
                if str(i[col_]) == str(val_):
                    db_[table]['datas'].remove(db_[table]['datas'][idx])
                idx += 1
        elif ' contain ' in where:
            col_ = where.split(' contain ')[0]
            val_ = where.split(' contain ')[1]
            idx = 0
            for i in db_[table]['datas']:
                if str(val_) in str(i[col_]):
                    db_[table]['datas'].remove(db_[table]['datas'][idx])
                idx += 1
            
        self.save_table(table, db_)

    def alter_table_rename_column(self,table:str, col:str, to_:str):
        db = self.openDB()
        db[table]["column"][to_] = db[table]["column"][col]
        del db[table]["column"][col]
        for i in reversed(db[table]["datas"]):
            data = i
            data[to_] = data[col]
            db[table]["datas"].append(data)
            db[table]["datas"].remove(i)
            del data[col]
        self.save_table(table, db)
        return
                
    def alter_table_add_column(self, table:str, col_name:str, datatypes:str):
        db = self.openDB()
        db[table]['column'][col_name] = datatypes
        for i in db[table]['datas']:
            if datatypes == 'str':
                i[col_name] = "null"
            elif datatypes == 'int':
                i[col_name] = 0
            elif datatypes == 'float':
                i[col_name] = 0.0
            elif datatypes == 'dict':
                i[col_name] = {}
            else:
                sys.exit("Unknown datatypes")
        self.save_table(table, db)
        return
    
    def alter_table_drop_column(self, table:str, col_name:str):
        db = self.openDB()
        if col_name not in db[table]['column']:
            sys.exit("Column not found")
        del db[table]['column'][col_name]
        for i in db[table]['datas']:
            del i[col_name]
        self.save_table(table, db)
        return
    
    def copy_table(self, table:str, to_table:str):
        db = self.openDB()
        if table not in db or to_table not in db:
            sys.exit("Table not found")
        if db[table]['column'] != db[to_table]['column']:
            sys.exit("The table structure must be same")
        db[to_table]['datas'] = db[table]['datas'] + db[to_table]['datas']
        self.save_table(to_table, db)
        return
    
    def clear_table(self, table:str):
        db = self.openDB()
        db[table]['datas'] = []
        self.save_table(table, db)
        return
    
    def count_rows(self, table) -> int:
        db = self.openDB()
        return len(db[table]['datas'])


import setuptools

setuptools.setup(
    name="sayuDB",
    version="0.0.6",
    author="Arsybai",
    description="Database management system based on python and JSON.",
    packages=["sayuDB"],
    license="MIT",
    author_email="me@arsybai.com",
    url="https://github.com/Arsybai/sayuDB",
    keywords=[
        'database',
    ],
    install_requires=[
    'requests',
    'tabulate',
    'flask'
    ],
    long_description_content_type="text/markdown",
    long_description="""# Documentation

##### Table of content
###### Preface
1. [What is sayuDB](#1_What_is_sayuDB_18)

###### Tutorial
1. [Getting Started](#1-getting-started)
- 1.1 [Installation](#11-installation)
- 1.2 [Creating database](#12-creating-database)
- 1.3 [Creating a New Table](#13-creating-a-new-table)
- 1.4 [Populating a table with rows](#14-pupulating-a-table-with-rows)
- 1.5 [Querying a Table](#15-querying-a-table)
- 1.6 [Updates](#updates)
- 1.7 [Deletion](#17-deletion)

---
# Preface
### 1. What is sayuDB
sayuDB is an database management system based on python and JSON. Developed at Clee Ltd. This project actually for personal purpose only but for some reason I publish it.
It supports a large part of the SQL standard feature

# Tutorial
### 1. Getting Started
#### 1.1 Installation
Before you can use sayuDB you need to install it, of course.
Just clone this repository and place it to your project folder.
```shell
> pip3 install sayuDB
```
for help menu:
```shell
> python3 sayuDB --h
```

#### 1.2 Creating database
The first test to see whether you can access the database server is to try to create a database.

To create a new database, in this example named `myDB`, you use the following funtion:
```python
import sayuDB

# Creating database
sayuDB.create_database('myDB')
```
You can also create databases with other names. sayuDB allows you to create any number of databases at a given site. Database names must have an alphabetic first character, can not contain space and are limited to 63 bytes in length.

If you do not want to use your database anymore you can remove it. For example, you can destroy it using the following function:
```python
import sayuDB

sayuDB.drop_database('myDB')
```

You can import export using:
```python
# Exporting database (must use file name)
sayuDB.export_database('<name_of_database>', path_='<path>/filename.ezdb')
# Importing database
sayuDB.import_database(path_='/<path>/filename.ezdb')
```

#### 1.3 Creating a New Table
You can create a new table by specifying the table name, along with all column names and their types:

you must define the database first with:
```python
import sayuDB

db = sayuDB.sayuDB(database='myDB')
```

then
```python
db.create_table('people', [
    ['name','str'],
    ['age','int'],
    ['city','str']
])
```
sayuDB currenly support data types `str`,`int`,`float`,`dict`

If you want to show the table, using this following function:
```python
db.show_table(name='people')
```
it will returned as printed table:
```shell
name    column    datas
------  --------  -------
```

Finally, it should be mentioned that if you don't need a table any longer or want to recreate it differently you can remove it using the following function:
```python
db.drop_table(name='people')
```

#### 1.4 Populating a Table With Rows
The `insert_row` function is used to populate a table with rows:
```python
db.insert_row(table='people', col='name,age,city', contents=['Arsybai', 23, 'Solo, Indonesia'])
```
_You can also use col as list. EX `['name','age','city']`_

### 1.5 Querying a Table
To retrieve data from a table, the table is queried. An `select_row` funtion is used to do this. The funtion is divided into a select list (the part that lists the columns to be returned), a table list (the part that lists the tables from which to retrieve the data), and an optional qualification (the part that specifies any restrictions). For example, to retrieve all the rows of table `people`, type:
```python
data = db.select_row(table='people', col='*')
print(data)
```
Here `*` is a shorthand for “all columns”. So the same result would be had with:
```python
db.select_row(table='people', col='name,age,city')
```
The output should be:
```shell
name       age  city
-------  -----  ------------------
Arsybai     23  Solo, Indonesia
Ataro       25  Bekasi, Indonesia
Moepoi      21  Jakarta, Indonesia
```

If you want the output is json, use `as_json=True`:
```python
data = db.select_row(table='people', col='*', as_json=True)
print(data)
```

A query can be “qualified” by adding a `where` clause that specifies which rows are wanted:
```python
data = db.select_row(table='people', col='*', where='age=23')
print(data)
```
you can also use ` contain ` for specific column that contain some value:
```python
data = db.select_row(table='people', col='*', where='age contain 23')
print(data)
```
Result:
```shell
name       age  city
-------  -----  ---------------
Arsybai     23  Solo, Indonesia
```

You can request that the results of a query be returned in sorted order:
```python
data = db.select_row(table='people', col='*', order_by='age|asc')
print(data)
```
_The order should be `asc` or `desc`_

Result:
```shell
name       age  city
-------  -----  ------------------
Moepoi      21  Jakarta, Indonesia
Arsybai     23  Solo, Indonesia
Ataro       25  Bekasi, Indonesia
```

#### Updates
You can update existing rows using the `update_row`.
as example I want to set `Ataro` age to `26`:
```python
db.update_row(table='people', set_='age=26', where='name=Ataro')
```
Look at the new state of the data:
```shell
name       age  city
-------  -----  ------------------
Arsybai     23  Solo, Indonesia
Ataro       26  Bekasi, Indonesia
Moepoi      21  Jakarta, Indonesia
```

#### Deletion
Rows can be removed from a table using the `delete_row`:
```python
db.delete_row(table='people', where='name=Ataro')
```

Look at the new state of the data:
```shell
name       age  city
-------  -----  ------------------
Arsybai     23  Solo, Indonesia
Moepoi      21  Jakarta, Indonesia
```
"""
)